/* refentry2man
 * Copyright (C) 2000 Rafa Kleger-Rudomin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


#include <stdio.h>
#include "convgen.h"
extern int multiple_files_option;
extern int be_quiet_option;
extern int debug_option;

enum SimpleListTypes { VERT, HORIZ, INLINE };
enum FontStyles { NO_FONT, R, B, I };
enum ListsTypes {
  ITEMIZEDLIST,
  ORDEREDLIST,
  VARIABLELIST,
  ADMONITION,
  PARA,
  GROUP,				/* group of args in cmdsynopsis */
  ARG
};

static XML_Char *lang = "";	/* refentry language (lang='?') */

TempBuffer refentrytitle;
TempBuffer manvolnum;
TempBuffer date;
TempBuffer source;
TempBuffer manual;

static int member_count;	/* needed to separate members of simplelist with commas */
static int simplelist_type;	/* VERT HORIZ INLINE */
static int refname_count = 0;	/* needed to separate refnames with commas */
static int paramdef_count = 0;	/* needed to separate function params with commas */
static int arg_optional = 0;	/* put braces around command argument? */
static int arg_repeat = 0;	/* put elipsis after argument? */
static int linespecific = 0;	/* vebatim environment? */
static int currentfont = NO_FONT;
static int fontbuffer = NO_FONT;

enum SpacebufferValues {
  NORMAL_CHAR,
  SPACE,	
  SUPPRESS_SPACES,		/* all spaces are ignored (start or end of block element) */
  /* verical breaks start here */
  NEWLINE,			/* \n */
  DOT_BR,			/* .br */
  DOT_SP,			/* .sp */
  DOT_P,			/* .P */
};
static int spacebuffer = SUPPRESS_SPACES;

static int in_citerefentry = 0;	/* manvolnum and refentytitle must be treated
				   differently depending on the element
				   they are included in
				 */

#define MAXLISTDEPTH 50
static int listlevel = 0;	/* current list nesting level */
static int listtype[MAXLISTDEPTH]; /* stack for types of nested list*/
static int listcounter[MAXLISTDEPTH]; /* stack for numberedlists' counters */
static int listindent[MAXLISTDEPTH]; /* stack for relative indents */


#define ITEMIZEDLIST_INDENT 2
#define ORDEREDLIST_INDENT 4
#define VARIABLELIST_INDENT 8
#define ADMONITION_INDENT 3


/* *****************************************************************/
/* ********************* utility functions *************************/
/* *****************************************************************/

void len_printf(const XML_Char *string, int len) {
  int i;
  char s;
  //printf("%d",linespecific);
  for(i = 0; i<len; i++) {
    s = *string;
    switch (s) {
    case ' ': case '\t': case '\n':
      if (linespecific) {
	if (s=='\t') printf("    ");
	else if (s=='\n') printf("\n.br\n");
	else putchar(' ');
      }
      else {
	/* don't print multiple whitespaces */
	if (spacebuffer == NORMAL_CHAR) {
	  spacebuffer = SPACE;
	}
      }
      break;
    default:
      /* any else than whitespace */
      /* first check if there is some space buffered */
      if (spacebuffer == SPACE) {
	putchar(' ');
      }
      else if (spacebuffer >= NEWLINE) {
	switch (spacebuffer) {
	  //case DOT_P: printf("\n.P\n"); break;
	  //case DOT_SP: printf("\n.sp\n"); break;
	  //case DOT_IP: break;
	case DOT_BR: printf("\n.br\n"); break;
	case NEWLINE: printf("\n"); break;
	default:
	  if (listlevel) printf("\n.IP \"\" %d\n", listindent[listlevel]);
	  else printf("\n.P\n");
	}
      }
      if (fontbuffer != NO_FONT) {
	switch (fontbuffer) {
	case R: printf("\\fR"); break;
	case B: printf("\\fB"); break;
	case I: printf("\\fI"); break;
	}
	fontbuffer = NO_FONT;
      }
      spacebuffer = NORMAL_CHAR;
      switch (s) {
      case '-':	printf("\\-");	break;
	//case '|':	printf("\\|");	break;
      case '\\': printf("\\ "); break; /* TODO */
      default: putchar(s);
      }
    }
    string++;
  }
}


const XML_Char *attr_value(const XML_Char **attr, const XML_Char *name) {
  int i;
  for (i = 0; attr[i]; i += 2) {
    if (strcmp(attr[i],name) == 0) return attr[i+1];
  }
  return "";
}

void font_change(int font) {
  if (font == currentfont) return;
  /* if (spacebuffer ) maybe flushing space will not be necessary */
  /*
  switch (font) {
  case R: printf("\\fR"); break;
  case B: printf("\\fB"); break;
  case I: printf("\\fI"); break;
  }
  */
  fontbuffer = font;
  currentfont = font;
}

void flush_space() {
  if (spacebuffer == SPACE) {
    putchar(' ');
    spacebuffer = NORMAL_CHAR;
  }
}

/* this changes font and prints data */
#define FONT(FF) {font_change(FF); len_printf(s, len); }
#define BLOCK_MIXED spacebuffer = SUPPRESS_SPACES; currentfont = NO_FONT;
#define BLOCK BLOCK_MIXED


/* *****************************************************************/
/* ************************** root element *************************/
/* *****************************************************************/
START(refentry)
{
  /*
    if(!be_quiet_option) fprintf(stderr,"refentry language found: lang='%s'\n", attr_value(attr,"lang"));
  */
  lang = strdup(attr_value(attr,"lang"));

  if (multiple_files_option) {
    refname_count = 0;	
    paramdef_count = 0;	
    arg_optional = 0;	
    arg_repeat = 0;	
    linespecific = 0;
    currentfont = NO_FONT;
    spacebuffer = SUPPRESS_SPACES;
    listlevel = 0;
    in_citerefentry = 0;
    refentrytitle.str = "";
    manvolnum.str = "";
    date.str = "";
    source.str = "";
    manual.str = "";
  }
  
}
END(refentry)
{
  /* fprintf(stderr,__FUNCTION__"\n"); */

  /* ignore input until next refentry */
  if (multiple_files_option) {
    free(lang);
    wait_for_start_tag("refentry");
  }
  else {
    exit(0);			/* one refentry is enough */
  }
}


/* *****************************************************************/
/* ************************ refmeta tags ***************************/
/* *****************************************************************/

START(refmeta) {
  BLOCK_MIXED;
}  
END(refmeta) {
  /* now we know enough to put output to file */
  char *tmp;
  const char *filename, *ext;
  int fd;
  BLOCK_MIXED;
  if (multiple_files_option) {
    filename = refentrytitle.str;
    ext = manvolnum.str;
    /* hack to remove \fB etc, if they got here */
    if (*filename == '\\') filename += 3;
    if (*ext == '\\') ext += 3;
    if ((*filename == '\0') || (*ext == '\0')) {
      fprintf(stderr, "incomplete 'refmeta' element found (lack of 'refentrytitle' or 'manvolnum') \n");
      //freopen("/dev/null", "w", stdout);
      exit(1);
    }
    tmp = (char *) malloc((strlen(filename)+strlen(ext)+2) * sizeof(char));
    sprintf(tmp, "%s.%s",filename, ext);
    /*
      if (freopen(tmp, "w", stdout) == NULL) {
      fprintf(stderr, "cannot freopen() file %s\n",tmp);
      exit(1);
      }
      */
    fflush(stdout);
    //fd = fileno(stdout);
    if ((stdout = fopen(tmp,"w")) != NULL) {
      fprintf(stderr, "creating file %s\n",tmp);
      //if (close(fd) != 0) {
      //  fprintf(stderr, "cannot close() current file \n");
      //  exit(1);
      //}
    }
    else {
      fprintf(stderr, "cannot fopen() file %s\n",tmp);
      exit(1);
    }
    
    free(tmp);
  }
  printf(".TH \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"",
	 refentrytitle.len ? refentrytitle.str : "",
	 manvolnum.len ? manvolnum.str : "",
	 date.len ? date.str : "",
	 source.len ? source.str : "",
	 manual.len ? manual.str : "");
}

START(refentrytitle) {
  if (!in_citerefentry) { set_output_to_temp_buffer(&refentrytitle); }
} 
END(refentrytitle) {
  if (!in_citerefentry) { unset_output_to_temp_buffer(); }
}
DATA(refentrytitle) {
  if (in_citerefentry) { FONT(B); } else { len_printf(s,len); }
}
 
START(manvolnum) {
  if (in_citerefentry) { font_change(R); len_printf("(",1); }
  else { set_output_to_temp_buffer(&manvolnum); }
} 
END(manvolnum) {
  if (in_citerefentry) { putchar(')'); }
  else { unset_output_to_temp_buffer(); }
} 
DATA(manvolnum) {
  if (in_citerefentry) { FONT(R); } else { len_printf(s,len); }
} 

START(refmiscinfo) {
  const XML_Char *v;
  v = attr_value(attr,"class");
  if (strcmp(v,"date")   == 0) { set_output_to_temp_buffer(&date);   }
  else if (strcmp(v,"source") == 0 || strcmp(v,"software") == 0)
    { set_output_to_temp_buffer(&source); }
  else if (strcmp(v,"manual") == 0 || strcmp(v,"sectdesc") == 0)
    { set_output_to_temp_buffer(&manual); }
  /* temporary hack for libggi */
  else { set_output_to_temp_buffer(&source); }
} 
END(refmiscinfo) { unset_output_to_temp_buffer(); }
DATA(refmiscinfo) FONT(R);

/* *****************************************************************/
/* ************************ refnamediv tags ************************/
/* *****************************************************************/

START(refnamediv) { BLOCK_MIXED; /*printf("\n");*/ }
END(refnamediv) { BLOCK_MIXED; /*printf("\n");*/ }

START(refname) { if (refname_count++) printf(", "); else printf("\n.SH NAME\n"); }
DATA(refname) { len_printf(s,len); }

START(refpurpose) { printf(" \\- "); }
DATA(refpurpose) { len_printf(s,len); }

/* *****************************************************************/
/* *********************** refsynopsisdiv tags *********************/
/* *****************************************************************/
START(refsynopsisdiv) { printf("\n.SH SYNOPSIS"); }

START(cmdsynopsis) { BLOCK_MIXED; printf("\n.br\n"); }
END(cmdsynopsis) { BLOCK_MIXED; }
DATA(cmdsynopsis) FONT(R);


/*
  args and groups are treated as lists
  listcounter keeps number of children here
  listindent keeps some status info here
*/

#define ARG_OPT 1
#define ARG_REQ 2
#define ARG_REPEAT 4

void start_arg(const XML_Char **attr, int type) {
  const XML_Char *a;
  int choice, repeat;

  choice = ARG_OPT;		/* default */
  repeat = 0;		/* default */
  /* if parent is <group>, must place vertical bars */
  if (listlevel)
    if (listtype[listlevel] == GROUP) {
      choice = 0;	/* in group the default is plain */
      /* listcounter counts child elements now */
      if (listcounter[listlevel]++ > 0) {
	if(spacebuffer == SPACE) { printf(" "); } /* if "1 | 2" not "1|2"*/
	printf("|");
      }
    }

  
  a = attr_value(attr,"rep");
  repeat = (strcmp(a,"repeat")==0) ? ARG_REPEAT : 0;
  a = attr_value(attr,"choice");
  if (strcmp(a,"plain")==0) { choice = 0; }
  else if (strcmp(a,"opt")==0) { choice = ARG_OPT; }
  else if (strcmp(a,"req")==0) { choice = ARG_REQ; }

  if (choice == ARG_OPT) { font_change(R); len_printf("[",1); }
  else if (choice == ARG_REQ) { font_change(R); len_printf("{",1); }

  /* register current element */
  ++listlevel;
  listtype[listlevel] = type;
  listcounter[listlevel] = 0;	/* listcounter is used to count args */
  listindent[listlevel] = (choice | repeat);
}
void end_arg(int type) {
  int status;
  status = listindent[listlevel];
  if (status & ARG_REPEAT) { font_change(R); len_printf("...",3); }
  if (status & ARG_OPT) { font_change(R); len_printf("]",1); }
  else if(status & ARG_REQ) { font_change(R); len_printf("}",1); }
  listlevel--;
}

START(arg) { start_arg(attr, ARG); }
END(arg) { end_arg(ARG); }
DATA(arg) FONT(B);

START(group) { start_arg(attr, GROUP); }
END(group) { end_arg(GROUP); }
DATA(group) FONT(B);

START(sbr) { BLOCK; printf("\n.br\n"); }	/* line break in cmdsynopsis */



START(funcsynopsis) { BLOCK_MIXED; /*printf("\n.sp\n");*/ }
END(funcsynopsis) { BLOCK_MIXED;/* printf("\n");*/ }

START(funcsynopsisinfo) { BLOCK_MIXED; linespecific = 1; printf("\n.P\n"); }
END(funcsynopsisinfo) { BLOCK_MIXED; linespecific = 0; /* printf("\n");*/ }
DATA(funcsynopsisinfo) FONT(B)

START(funcprototype) { BLOCK_MIXED; printf("\n.HP\n"); }
END(funcprototype) { BLOCK_MIXED; font_change(B); printf(")"); }	/* closing paren */

END(funcdef) { paramdef_count = 0; font_change(B); printf("("); } /* opening paren */
DATA(funcdef) FONT(B)
     
START(paramdef) { if(paramdef_count++) printf(", "); }
DATA(paramdef) FONT(B)
     

START(funcparams) { font_change(B); printf("("); }
END(funcparams) { font_change(B); printf(")"); }
DATA(funcparams) FONT(I)

START(void) { printf("void"); }
START(varargs) { if(paramdef_count++) printf(", "); font_change(I); printf("\\&..."); }


START(synopsis) { BLOCK_MIXED; linespecific = 1; printf("\n.sp\n"); }
END(synopsis) { BLOCK_MIXED; linespecific = 0; /* printf("\n");*/ }
DATA(synopsis) FONT(R)

START(optional) { font_change(R); printf(" [ "); }
END(optional) { font_change(R); printf(" ] "); }
DATA(optional) FONT(B)

     
/* refsect* */
START(refsect1) { BLOCK; printf("\n.SH "); }
START(refsect2) { BLOCK; printf("\n.SS "); }
START(refsect3) { BLOCK; printf("\n.SS \" \""); } /* ugly */



/* *****************************************************************/
/* ************************** block elements ***********************/
/* *****************************************************************/

/* function for lists */
void start_list(int type, int indent) {
  ++listlevel;
  listtype[listlevel] = type;
  listcounter[listlevel] = 0;
  if (listlevel > 1) printf("\n.RS %d",listindent[listlevel-1]);
  else { printf("\n.RS 0"); }		/* this in not a bug! */
  listindent[listlevel] = indent;
}
void end_list() {
  /* if (listlevel > 1)*/
  printf("\n.RE");
  listlevel--;
}

/* function for paragraphs */
void start_para() {
  //BLOCK_MIXED; /* if(!listlevel) printf("\n.P\n"); else printf("\n");*/
  /*  
      ++listlevel;
      listtype[listlevel] = PARA;
      if (spacebuffer <= DOT_P) {
      if (listlevel > 1) printf("\n.RS %d",listindent[listlevel-1]);
      listindent[listlevel] = 0;
      spacebuffer = DOT_SP;
      }
  */
  
  /*
    fprintf(stderr,"XXXXXXXXXXX %d\n", spacebuffer);
    if (spacebuffer <= DOT_SP) { spacebuffer = DOT_SP; }
    else if (spacebuffer == DOT_P) {}
    else { spacebuffer = NEWLINE; }
  */
  /*
    spacebuffer = DOT_SP;
    currentfont = NO_FONT;
  */
  // spacebuffer = DOT_SP;
}
void end_para() {
  //BLOCK_MIXED; /* if(listlevel) printf("\n.br"); */
  /*
    if (listlevel > 1) printf("\n.RE");
    listlevel--;
  */
  spacebuffer = DOT_SP;
  currentfont = NO_FONT;
}
     
START(para) { start_para(); }
END(para) { end_para(); }
DATA(para) FONT(R);

START(formalpara) { /* just a wrapper */ }

START(simpara) { start_para(); }
END(simpara) { end_para(); }
DATA(simpara) FONT(R);

START(highlights) {
  printf("\n.RS +%d",ADMONITION_INDENT);
  start_para(); }
END(highlights) {
  end_para();
  printf("\n.RE");
}

START(simplelist) {
  const XML_Char *a;
  BLOCK;
  member_count = 0;
  a = attr_value(attr,"type");
  /* columns not implemented so vert and horiz list looks the same */
  if (strcmp(a,"inline")==0) {
    simplelist_type = INLINE;
  }
  /*else if (strcmp(a,"horiz")==0) {
    simplelist_type = HORIZ;
    } */
  else {
    simplelist_type = VERT;
  }
}

START(member) {
  switch (simplelist_type) {
  case INLINE:
    if(member_count++ > 0) { printf(", "); } else { printf(" "); }
    break;
  case VERT:
  case HORIZ:
    printf("\n.br\n");
    break;
  }
}
DATA(member) FONT(R);

START(example) {
  BLOCK;
  if (listlevel) printf("\n.IP \"\" %d\n", listindent[listlevel]);
  else printf("\n.P\n");
  font_change(I); len_printf("Example:",8);
  spacebuffer = SPACE;
}
END(example) {
  spacebuffer = DOT_P;
}

START(itemizedlist) {
  BLOCK;
  start_list(ITEMIZEDLIST, ITEMIZEDLIST_INDENT);
}
END(itemizedlist) {
  BLOCK;
  end_list();
  spacebuffer = DOT_P;
}
START(orderedlist) {
  BLOCK;
  start_list(ORDEREDLIST, ORDEREDLIST_INDENT);
}
END(orderedlist) {
  BLOCK;
  end_list();
  spacebuffer = DOT_P;
}
START(variablelist) {
  BLOCK;
  start_list(VARIABLELIST, VARIABLELIST_INDENT);
}
END(variablelist) {
  BLOCK;
  end_list();
  spacebuffer = DOT_P;
}


START(varlistentry) {
  printf("\n.IP \"");
  BLOCK;
  /* rest is in <term> element */
}



START(listitem) {
  switch (listtype[listlevel]) {
  case ITEMIZEDLIST:
    printf("\n.IP \\(bu %d", ITEMIZEDLIST_INDENT);
    spacebuffer = NEWLINE;
    break;
  case ORDEREDLIST:
    printf("\n.IP \"%2d.\" %d",++listcounter[listlevel], ORDEREDLIST_INDENT);
    spacebuffer = NEWLINE;
    break;
  case VARIABLELIST:
    printf("\" %d", VARIABLELIST_INDENT);
    spacebuffer = NEWLINE;
    break;

  }
}

START(screen) { BLOCK_MIXED linespecific = 1; printf("\n.na\n");}
END(screen) { BLOCK_MIXED linespecific = 0; printf("\n.ad"); spacebuffer = DOT_P;}
DATA(screen) FONT(R)

START(programlisting) { BLOCK_MIXED linespecific = 1; printf("\n.na\n");}
END(programlisting) { BLOCK_MIXED linespecific = 0; printf("\n.ad"); spacebuffer = DOT_P;}
DATA(programlisting) FONT(R)

START(literallayout) { BLOCK_MIXED linespecific = 1; printf("\n.na\n");}
END(literallayout) { BLOCK_MIXED linespecific = 0; printf("\n.ad"); spacebuffer = DOT_P;}
DATA(literallayout) FONT(R)
     
DATA(term) FONT(B);

/*
  UR causes problems
xxSTART(ulink) {
  const XML_Char *a;
  a = attr_value(attr,"url");
  printf("\n.UR \"");
  if (a) printf("%s",a);
  printf("\"\n");
}
xxEND(ulink) { printf("\n.UE\n"); }
*/

DATA(ulink) FONT(B);


START(quote) { /* flush_space();*/ len_printf("\"",1); }
END(quote) { printf("\""); }
DATA(quote) { len_printf(s, len); }

/* comments are replaced by remarks in db4.0 */
START(remark) { font_change(B); len_printf("[Remark:",8); spacebuffer = SPACE;}
END(remark) { font_change(B); printf("]"); }
DATA(remark) FONT(R);

START(comment) { font_change(B); len_printf("[Comment:",9); spacebuffer = SPACE;}
END(comment) { font_change(B); printf("]"); }
DATA(comment) FONT(R);


void start_admonition(const XML_Char *text, int len) {
  BLOCK;
  start_list(ADMONITION, ADMONITION_INDENT);
  printf("\n.ll -%d",ADMONITION_INDENT);
  if (listlevel) printf("\n.IP \"\" %d\n", listindent[listlevel]);
  else printf("\n.P\n");
  font_change(B); len_printf(text, len);
  spacebuffer = SPACE;
}

void end_admonition() {
  BLOCK; end_list();  printf("\n.ll +%d",ADMONITION_INDENT);
  spacebuffer = DOT_P;
}

START(warning) { start_admonition("Warning!",8); }
END(warning) { end_admonition(); }
START(caution) { start_admonition("Caution!",8); }
END(caution) { end_admonition(); }
START(note) { start_admonition("Note:",5); }
END(note) { end_admonition(); }
START(tip) { start_admonition("Tip:",4); }
END(tip) { end_admonition(); }


START(citerefentry) { currentfont = NO_FONT; in_citerefentry = 1; }
END(citerefentry) { in_citerefentry = 0; }


START(abstract) { start_admonition("Abstract:",9); }
END(abstract) { end_admonition(); }

END(title) { spacebuffer = NEWLINE; currentfont = NO_FONT;}
DATA(title) FONT(B);

START(subtitle) { spacebuffer = SPACE; len_printf("-",1); spacebuffer = SPACE; }
END(subtitle) { spacebuffer = NEWLINE; currentfont = NO_FONT;}
DATA(subtitle) FONT(R);






/* *****************************************************************/
/* ********************* some inline elements **********************/
/* *****************************************************************/

DATA(author) FONT(R)
DATA(acronym) FONT(B)		/* should be small (SM) */
DATA(application) FONT(B)
DATA(command) FONT(B)
DATA(email) FONT(I)
DATA(emphasis) FONT(I)
DATA(envar) FONT(B)
DATA(errorcode) FONT(B)
DATA(firstname) FONT(R)
DATA(function) FONT(B)
DATA(foreignphrase) FONT(I)
DATA(filename) FONT(I)
DATA(literal) FONT(B)
DATA(markup) FONT(B)
DATA(option) FONT(B)
DATA(parameter) FONT(I)
DATA(productname) FONT(B)
DATA(prompt) FONT(R)
DATA(property) FONT(I)
DATA(replaceable) FONT(I)
DATA(returnvalue) FONT(B)
DATA(symbol) FONT(B)
DATA(structfield) FONT(I)
DATA(structname) FONT(B)
DATA(surname) FONT(R)
DATA(token) FONT(I)
DATA(userinput) FONT(B)
DATA(varname) FONT(I)
DATA(type) FONT(B)


     
/* default handler - for all unspecified elements */
START(default) {
  if(debug_option) printf("<%s>",name);
  if(!be_quiet_option) fprintf(stderr,"%s element unsupported\n", name);
}
END(default) {
  if(debug_option) printf("</%s>",name);
}
DATA(default) { len_printf(s, len); }
